/*
   SHAPESHIFTER DIY OPEN SOURCE DRUM MACHINE
   by FASELUNARE.COM

   https://github.com/faselunare/shapeshifter
   Released under the GPLv3 license.

  Author: Arturo Guadalupi
*/

#include <Scheduler.h>
#include "structures.h"
#include "constants.h"
#include "src/LEDs/SSLEDs.h"
#include "src/Keypad/SSKeypad.h"
#include "src/Pattern/SSPattern.h"
#include "src/Audio/SSAudio.h"
#include "src/Display/SSDisplay.h"
#include "src/Memory/SSMemory.h"

//#define DEBUG

trackInfo tracks[NUMBER_OF_TRACKS];
uint8_t currentTrack = TRACK_A, previousTrack = TRACK_A;
uint8_t previousSample = 1, previousLength = 16, previousNoteSpeed = 1, previousPattern[NUMBER_OF_TRACKS] = {0};

bool restoreView = false;

uint16_t previousTempo, currentTempo;

uint8_t functionNumber = NO_FUNCTION;

enum memoryState { noneState, saveState, loadState, doneState };
uint8_t memorySlot = 1;

uint8_t oldClkStep = 0;

memoryState memoryProgress = noneState;

void setup() {
  
  //Initialize structure
  for (int i = 0; i < NUMBER_OF_TRACKS; i++) {
    tracks[i].currentPattern = 0;
    previousPattern[i] = 0;
    for (int j = 0; j < NUMBER_OF_SUBPATTERNS; j++) {
      tracks[i].pattern[j] = 0;
      tracks[i].length[j] = NUMBER_OF_STEPS;
      tracks[i].sample[j] = 0;
    }
  }

  Serial.begin(115200);
  SSLEDs.begin(25); //brightness 25
  SSAudio.begin();
  SSKeypad.begin();
  SSDisplay.begin();
  delay(2000);

  SSLEDs.setFunctionColor(FUNCTION_A, 255, 0, 0);         //RED
  SSLEDs.setFunctionColor(FUNCTION_B, 0, 255, 0);         //GREEN
  SSLEDs.setFunctionColor(FUNCTION_C, 0, 0, 255);         //BLUE
  SSLEDs.setFunctionColor(FUNCTION_D, 255, 0, 255);       //PURPLE
  SSLEDs.setFunctionColor(FUNCTION_E, 255, 255, 0);       //ORANGE
  SSLEDs.setFunctionColor(FUNCTION_PLAY, 255, 255, 255);  //WHITE
  SSLEDs.setFunctionColor(FUNCTION_SAVE, 127, 255, 0);    //
  SSLEDs.setFunctionColor(FUNCTION_LOAD, 127, 0, 255);    //

  SSLEDs.setTrackColor(1, 255, 0, 0);         //RED
  SSLEDs.setTrackColor(2, 0, 255, 0);         //GREEN
  SSLEDs.setTrackColor(3, 0, 0, 255);         //BLUE
  SSLEDs.setTrackColor(4, 255, 0, 255);       //PURPLE
  SSLEDs.setTrackColor(5, 127, 0, 0);         //LIGHT RED
  SSLEDs.setTrackColor(6, 0, 127, 0);         //LIGHT GREEN
  SSLEDs.setTrackColor(7, 0, 0, 127);         //LIGHT BLUE
  SSLEDs.setTrackColor(8, 127, 0, 127);       //LIGHT PURPLE

  //SSLEDs.showLengths(tracks);

  pinMode(MASTERRST_PIN, INPUT_PULLUP);
  pinMode(CLKOUT_PIN, OUTPUT);
  pinMode(CLKIN_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(MASTERRST_PIN), resetCount, FALLING);

  Scheduler.startLoop(ledScan);
  Scheduler.startLoop(readKeypadLoop);
  Scheduler.startLoop(memoryManagement);
  Scheduler.startLoop(clkOut);
}

void resetCount() {
  SSAudio.resetCount();
  oldClkStep = 0;
}

void loop() {
  SSAudio.run(tracks, currentTrack);      //start audio

  if (SSDisplay.isConnected()) {
    switch (memoryProgress) {
      case saveState:
        SSDisplay.saving();
        break;

      case loadState:
        SSDisplay.loading();
        break;

      case doneState:
        SSDisplay.done();
        restoreView = true;
        break;

      default:
        currentTempo = SSAudio.getTempo();

        if (functionNumber == FUNCTION_A) {
          if ((!restoreView) || (currentTrack != previousTrack)) {
            previousTrack = currentTrack;
            SSDisplay.trackSelection(currentTrack);
            restoreView = true;
          }
        }
        else if (functionNumber == FUNCTION_B) {
          if ((!restoreView) || (tracks[currentTrack].sample[tracks[currentTrack].currentPattern] != previousSample)) {
            previousSample = tracks[currentTrack].sample[tracks[currentTrack].currentPattern];
            SSDisplay.sampleSelection(tracks[currentTrack].sample[tracks[currentTrack].currentPattern], currentTrack);
            restoreView = true;
          }
        }
        else if (functionNumber == FUNCTION_C) {
          if ((!restoreView) || (SSAudio.getNoteSpeed() != previousNoteSpeed)) {
            previousNoteSpeed = SSAudio.getNoteSpeed();
            SSDisplay.speedSelection(SSAudio.getNoteSpeed());
            restoreView = true;
          }
        }
        else if (functionNumber == FUNCTION_D) {
          if ((!restoreView) || (tracks[currentTrack].currentPattern != previousPattern[currentTrack])) {
            previousPattern[currentTrack] = tracks[currentTrack].currentPattern;
            SSDisplay.patternSelection(tracks[currentTrack].currentPattern + 1);
            restoreView = true;
          }
        }
        else if (functionNumber == FUNCTION_E) {
          if ((!restoreView) || (tracks[currentTrack].length[tracks[currentTrack].currentPattern] != previousLength)) {
            previousLength = tracks[currentTrack].length[tracks[currentTrack].currentPattern];
            SSDisplay.lengthSelection(tracks[currentTrack].length[tracks[currentTrack].currentPattern]);
            restoreView = true;
          }
        }
        else if ((currentTempo != previousTempo) || (restoreView)) {
          restoreView = false;
          previousTempo = currentTempo;
          SSDisplay.printInfo(currentTempo, currentTrack, tracks[currentTrack].sample[tracks[currentTrack].currentPattern], tracks[currentTrack].length[tracks[currentTrack].currentPattern], SSAudio.getNoteSpeed(), tracks[currentTrack].currentPattern + 1);
        }
        break;
    }
  }

  yield();
}

void ledScan() {
  SSLEDs.scan(currentTrack, functionNumber);

  yield();
}

void readKeypadLoop() {
  while (SSKeypad.available() > 0) { //Check for Interrupt flag and process
    keyEvent event = SSKeypad.readKeys();
#ifdef DEBUG
    Serial.print("Key ");
    Serial.print(event.value);
#endif
    if (event.status) {
#ifdef DEBUG
      Serial.println(" is DOWN");
#endif

      switch (event.value) {
        default:

          switch (functionNumber) {
            case FUNCTION_A:
              currentTrack = event.value;

              if (currentTrack >= NUMBER_OF_TRACKS - 1) {
                currentTrack = NUMBER_OF_TRACKS - 1;
              }
              SSPattern.set(tracks[currentTrack].pattern[tracks[currentTrack].currentPattern]);
              SSLEDs.patternOn(currentTrack, SSPattern.getPattern()); //show LEDs
              break;

            case FUNCTION_B:
              tracks[currentTrack].sample[tracks[currentTrack].currentPattern] = event.value;
#ifdef DEBUG
              Serial.print("New sample: ");
              Serial.print(tracks[currentTrack].sample[tracks[currentTrack].currentPattern]);
#endif
              break;

            case FUNCTION_C:
              switch (event.value) {
                case 0: SSAudio.setNoteSpeed(1); break;

                case 1: SSAudio.setNoteSpeed(2); break;

                case 2: SSAudio.setNoteSpeed(4); break;

                default: break;
              }
#ifdef DEBUG
              Serial.print("New speed: ");
              Serial.print(SSAudio.getNoteSpeed());
#endif
              break;

            case FUNCTION_D:
              tracks[currentTrack].currentPattern = event.value;

              if (tracks[currentTrack].currentPattern >= NUMBER_OF_SUBPATTERNS) {
                tracks[currentTrack].currentPattern = NUMBER_OF_SUBPATTERNS - 1;
              }

              SSPattern.set(tracks[currentTrack].pattern[tracks[currentTrack].currentPattern]);
              SSLEDs.patternOn(currentTrack, SSPattern.getPattern()); //show LEDs

#ifdef DEBUG
              Serial.print("New pattern: ");
              Serial.print(tracks[currentTrack].currentPattern);
#endif
              break;

            case FUNCTION_E:
              tracks[currentTrack].length[tracks[currentTrack].currentPattern] = event.value + 1;
              SSLEDs.lengthOn(currentTrack, tracks[currentTrack].length[tracks[currentTrack].currentPattern]);
              //SSLEDs.showLengths(tracks);
#ifdef DEBUG
              Serial.print("New length: ");
              Serial.println(tracks[currentTrack].length[tracks[currentTrack].currentPattern]);
#endif
              break;

            case FUNCTION_LOAD: case FUNCTION_SAVE:
              memorySlot = event.value;
              if (functionNumber == FUNCTION_LOAD) {
                memoryProgress = loadState;
              }
              else {
                memoryProgress = saveState;
              }

              break;

            default:
              SSPattern.update(event.value); //flip the bit
              tracks[currentTrack].pattern[tracks[currentTrack].currentPattern] = SSPattern.getPattern(); //update local pattern
              SSLEDs.patternOn(currentTrack, SSPattern.getPattern()); //show LEDs
              break;
          }
          break;

        case BUTTON_A: case BUTTON_B: case BUTTON_C: case BUTTON_D: case BUTTON_E:
          functionNumber = event.value - BUTTON_A;
          SSLEDs.functionOn(functionNumber);
          if (functionNumber == FUNCTION_E) {
            SSLEDs.lengthOn(currentTrack, tracks[currentTrack].length[tracks[currentTrack].currentPattern]);
          }
          break;

        case BUTTON_SAVE:
          functionNumber = FUNCTION_SAVE;
          SSLEDs.functionOn(functionNumber);
          break;

        case BUTTON_LOAD:
          functionNumber = FUNCTION_LOAD;
          SSLEDs.functionOn(functionNumber);
          break;

        case BUTTON_PLAY:
          //Pause playing
          if (SSAudio.isPlaying()) {
            SSAudio.pause();
          }
          else {
            SSAudio.play();
          }
          break;
      }
    }
    else {
#ifdef DEBUG
      Serial.println(" is UP");
#endif

      switch (event.value) {
        case BUTTON_A: case BUTTON_B: case BUTTON_C: case BUTTON_D: case BUTTON_E:
          functionNumber = event.value - BUTTON_A;
          SSLEDs.functionOff(functionNumber);
          if (functionNumber == FUNCTION_E) {
            SSLEDs.patternOn(currentTrack, SSPattern.getPattern()); //restore pattern view
          }
          functionNumber = NO_FUNCTION;
          break;

        case BUTTON_LOAD:
          functionNumber = NO_FUNCTION;
          SSLEDs.functionOff(FUNCTION_LOAD);
          break;

        case BUTTON_SAVE:
          functionNumber = NO_FUNCTION;
          SSLEDs.functionOff(FUNCTION_SAVE);
          break;

        default: break;
      }
    }
  }

  yield();
}

void memoryManagement() {
  switch (memoryProgress) {
    case saveState:
      if (SSAudio.isPlaying()) {
        SSAudio.pause();
        SSMemory.save(tracks, memorySlot);
        SSAudio.play();
      }
      else {
        SSMemory.save(tracks, memorySlot);
      }

      memoryProgress = doneState;
      break;

    case loadState:
      if (SSAudio.isPlaying()) {
        SSAudio.pause();
        SSMemory.load(tracks, memorySlot);
        SSAudio.play();
      }
      else {
        SSMemory.load(tracks, memorySlot);
      }
      memoryProgress = doneState;
      break;

    case doneState:
      delay(2000);
      memoryProgress = noneState;
      SSPattern.set(tracks[currentTrack].pattern[tracks[currentTrack].currentPattern]);
      SSLEDs.patternOn(currentTrack, SSPattern.getPattern()); //show LEDs
      restoreView = true;
      break;

    default: memoryProgress = noneState; break;
  }

  yield();
}

void clkOut() {
  if (oldClkStep == 0) {
    SSLEDs.on(MASTERRST_LED);
  }
  else {
    SSLEDs.off(MASTERRST_LED);
  }
  if ((SSAudio.isPlaying()) && (oldClkStep != SSAudio.getCurrentStep(TRACK_A))) {
    oldClkStep = SSAudio.getCurrentStep(TRACK_A);
    digitalWriteFast(CLKOUT_PIN, HIGH);
    delay(5);
    digitalWriteFast(CLKOUT_PIN, LOW);
  }
  if (oldClkStep % 2 == 0) {
    SSLEDs.on(CLKOUT_LED);
  } else {
    SSLEDs.off(CLKOUT_LED);
  }
  yield();
}
