/*
   SHAPESHIFTER DIY OPEN SOURCE DRUM MACHINE
   by FASELUNARE.COM

   https://github.com/faselunare/shapeshifter
   Released under the GPLv3 license.

  Author: Arturo Guadalupi
*/

#include <Wire.h>
#include "SSKeypad.h"

static const uint16_t DEFAULT_COLS = (COL0 | COL1 | COL2 | COL3 | COL4);
static const uint8_t DEFAULT_ROWS = (ROW0 | ROW1 | ROW2 | ROW3 | ROW4);

RingBufferN<KEY_COUNT_SIZE> keyEventBuffer;

void keyEventISR(void) {
  keyEventBuffer.store_char(SSKeypad.getKeyEvent());
  SSKeypad.clearInterruptStatus();
}

int SSKeypadClass::available() {
  return keyEventBuffer.available();
}

keyEvent SSKeypadClass::readKeys() {
  keyEvent key;
  uint8_t temp = keyEventBuffer.read_char();

  //Serial.print("CODE: ");
  //Serial.println(temp & 0x7F, HEX);

  uint8_t code = temp & 0x7F;

  switch (temp & 0x7F) {
    //Patterns
    case 0x01: key.value = 0; break;
    case 0x02: key.value = 1; break;
    case 0x03: key.value = 2; break;
    case 0x04: key.value = 3; break;
    case 0x05: key.value = 4; break;
    case 0x0B: key.value = 5; break;
    case 0x0C: key.value = 6; break;
    case 0x0D: key.value = 7; break;
    case 0x0E: key.value = 8; break;
    case 0x0F: key.value = 9; break;
    case 0x15: key.value = 10; break;
    case 0x16: key.value = 11; break;
    case 0x17: key.value = 12; break;
    case 0x18: key.value = 13; break;
    case 0x19: key.value = 14; break;
    case 0x1F: key.value = 15; break;

    //Tracks
    case 0x20: key.value = 16; break;
    case 0x21: key.value = 17; break;
    case 0x22: key.value = 18; break;
    case 0x23: key.value = 19; break;

    //Fn
    case 0x29: key.value = 20; break;

    //Play
    case 0x2A: key.value = 21; break;

    //Encoder
    case 0x2B: key.value = 22; break;

    //Save
    case 0x2D: key.value = 23; break;

    //Load
    case 0x2C: key.value = 24; break;

    default: break;
  }

  key.status = temp & 0x80;
  return key;
}

void SSKeypadClass::begin(uint8_t address, uint8_t IRQPin, uint8_t rstPin) {
  _address = address;
  _IRQPin = IRQPin;
  _rstPin = rstPin;
  // pinMode(rstPin, OUTPUT);
  // digitalWrite(rstPin, LOW);
  // delay(100);
  // digitalWrite(rstPin, HIGH);

  uint8_t config = CFG_KE_IEN | CFG_GPI_IEN  | CFG_OVR_FLOW_M | CFG_INT_CFG | CFG_GPI_E_CFG;
  attachInterrupt(digitalPinToInterrupt(IRQPin), keyEventISR);
  Wire.begin();
  configureKeys(DEFAULT_ROWS, DEFAULT_COLS, config);
}

char SSKeypadClass::readKeypad(void) {
  return getKeyEvent();
}

/*
  Configure the TCA8418 for keypad operation
*/
void SSKeypadClass::configureKeys(uint8_t rows, uint16_t cols, uint8_t config) {
  //Pins all default to GPIO. pinMode(x, KEYPAD); may be used for individual pins.
  writeByte(rows, REG_KP_GPIO1);

  uint8_t col_tmp;
  col_tmp = (uint8_t)(0xff & cols);
  writeByte(col_tmp, REG_KP_GPIO2);
  col_tmp = (uint8_t)(0x03 & (cols >> 8));
  writeByte(col_tmp, REG_KP_GPIO3);

  config |= CFG_AI;
  writeByte(config, REG_CFG);
}

void SSKeypadClass::writeByte(uint8_t data, uint8_t reg) {
  Wire.beginTransmission(_address);
  Wire.write((uint8_t) reg);

  Wire.write((uint8_t) data);
  Wire.endTransmission();
}

bool SSKeypadClass::readByte(uint8_t *data, uint8_t reg) {
  Wire.beginTransmission(_address);
  Wire.write((uint8_t) reg);
  Wire.endTransmission();
  uint8_t timeout = 0;

  Wire.requestFrom(_address, (uint8_t) 0x01);
  while (Wire.available() < 1) {
    timeout++;
    if (timeout > I2CTIMEOUT) {
      return (true);
    }
    delay(1);
  }       // Experimental

  *data = Wire.read();

  return (false);
}

void SSKeypadClass::write3Bytes(uint32_t data, uint8_t reg) {

  union
  {
    uint8_t b[4];
    uint32_t w;
  } datau;

  datau.w = data;

  Wire.beginTransmission(_address);
  Wire.write((uint8_t) reg);

  Wire.write((uint8_t) datau.b[0]);
  Wire.write((uint8_t) datau.b[1]);
  Wire.write((uint8_t) datau.b[2]);

  Serial.println((uint8_t) datau.b[0]);
  Serial.println((uint8_t) datau.b[1]);
  Serial.println((uint8_t) datau.b[2]);

  Wire.endTransmission();
  return;
}

bool SSKeypadClass::read3Bytes(uint32_t *data, uint8_t reg) {

  union
  {
    uint8_t b[4];
    uint32_t w;
  } datau;

  datau.w = *data;

  Wire.beginTransmission(_address);
  Wire.write((uint8_t) reg);
  Wire.endTransmission();
  uint8_t timeout = 0;

  Wire.requestFrom(_address, (uint8_t) 0x03);
  while (Wire.available() < 3) {
    timeout++;
    if (timeout > I2CTIMEOUT) {
      return (true);
    }
    delay(1);
  }     //Experimental

  datau.b[0] = Wire.read();
  datau.b[1] = Wire.read();
  datau.b[2] = Wire.read();

  Serial.println(datau.b[0]);
  Serial.println(datau.b[1]);
  Serial.println(datau.b[2]);

  *data = datau.w;

  return (false);
}

void SSKeypadClass::pinMode(uint32_t pin, uint8_t mode) {
  uint32_t dbc;

  readGPIO();
  switch (mode) {
    case INPUT:
      bitClear(_PORT, pin);
      bitClear(_DDR, pin);
      bitClear(_PKG, pin);
      bitSet(_PUR, pin);
      break;
    case INPUT_PULLUP:
      bitClear(_PORT, pin);
      bitClear(_DDR, pin);
      bitClear(_PKG, pin);
      bitClear(_PUR, pin);
      break;
    case OUTPUT:
      bitClear(_PORT, pin);
      bitSet(_DDR, pin);
      bitClear(_PKG, pin);
      bitSet(_PUR, pin);
      break;
    case KEYPAD:
      bitClear(_PORT, pin);
      bitClear(_DDR, pin);
      bitSet(_PKG, pin);
      bitClear(_PUR, pin);
      break;
    case DEBOUNCE:
      read3Bytes((uint32_t *)&dbc, REG_DEBOUNCE_DIS1);
      bitClear(dbc, pin);
      write3Bytes((uint32_t)dbc, REG_DEBOUNCE_DIS1);
      break;
    case NODEBOUNCE:
      read3Bytes((uint32_t *)&dbc, REG_DEBOUNCE_DIS1);
      bitSet(dbc, pin);
      write3Bytes((uint32_t)dbc, REG_DEBOUNCE_DIS1);
      break;
    default:
      break;
  }
  updateGPIO();
}


void SSKeypadClass::digitalWrite(uint32_t pin, uint8_t value) {
  if (value)
    bitSet(_PORT, pin);
  else
    bitClear(_PORT, pin);

  updateGPIO();
}

uint8_t SSKeypadClass::digitalRead(uint32_t pin) {
  readGPIO();

  return (_PIN & bit(pin)) ? HIGH : LOW;
}

void SSKeypadClass::write(uint32_t value) {
  _PORT = value;

  updateGPIO();
}

uint32_t SSKeypadClass::read(void) {
  readGPIO();

  return _PORT;
}

void SSKeypadClass::toggle(uint32_t pin) {
  _PORT ^= (bit(pin));

  updateGPIO();
}

void SSKeypadClass::blink(uint32_t pin, uint16_t count, uint32_t duration) {
  duration /= count * 2;

  while (count--) {
    toggle(pin);
    delay(duration);
    toggle(pin);
    delay(duration);
  }
}

void SSKeypadClass::attachInterrupt(int pin, void(*selfCheckFunction)(void)) {
  _IRQPin = pin;
  ::pinMode(_IRQPin, INPUT_PULLUP);
  ::attachInterrupt(digitalPinToInterrupt(_IRQPin), selfCheckFunction, FALLING);
}

void SSKeypadClass::detachInterrupt() {
  ::detachInterrupt(digitalPinToInterrupt(_IRQPin));
}

void SSKeypadClass::pinInterruptMode(uint32_t pin, uint8_t mode, uint8_t level, uint8_t fifo) {
  uint32_t intSetting, levelSetting, eventmodeSetting;

  read3Bytes((uint32_t *)&intSetting, REG_GPIO_INT_EN1);
  read3Bytes((uint32_t *)&levelSetting, REG_GPIO_INT_LVL1);
  read3Bytes((uint32_t *)&eventmodeSetting, REG_GPI_EM1);

  switch (mode) {
    case INTERRUPT:
      bitSet(intSetting, pin);
      break;
    case NOINTERRUPT:
      bitClear(intSetting, pin);
      break;
    default:
      break;
  }

  switch (level) {
    case LOW:
      bitClear(levelSetting, pin);
      break;
    case HIGH:
      bitSet(levelSetting, pin);
      break;
    default:
      break;
  }

  switch (fifo) {
    case FIFO:
      bitSet(eventmodeSetting, pin);
      break;
    case NOFIFO:
      bitClear(eventmodeSetting, pin);
      break;
    default:
      break;
  }

  write3Bytes((uint32_t)intSetting, REG_GPIO_INT_EN1);
  write3Bytes((uint32_t)levelSetting, REG_GPIO_INT_LVL1);
  write3Bytes((uint32_t)eventmodeSetting, REG_GPI_EM1);

}

void SSKeypadClass::pinInterruptMode(uint32_t pin, uint8_t mode) {
  pinInterruptMode(pin, mode, 0, 0);
}

void SSKeypadClass::readGPIO() {
  read3Bytes((uint32_t *)&_PORT, REG_GPIO_DAT_OUT1);  //Read Data OUT Registers
  read3Bytes((uint32_t *)&_PIN, REG_GPIO_DAT_STAT1);  //Read Data STATUS Registers
  read3Bytes((uint32_t *)&_DDR, REG_GPIO_DIR1);   //Read Data DIRECTION Registers
  read3Bytes((uint32_t *)&_PKG, REG_KP_GPIO1);    //Read Keypad/GPIO SELECTION Registers
  read3Bytes((uint32_t *)&_PUR, REG_GPIO_PULL1);    //Read KPull-Up RESISTOR Registers
}

void SSKeypadClass::updateGPIO() {
  write3Bytes((uint32_t)_PORT, REG_GPIO_DAT_OUT1);    //Write Data OUT Registers
  write3Bytes((uint32_t)_DDR, REG_GPIO_DIR1);     //Write Data DIRECTION Registers
  write3Bytes((uint32_t)_PKG, REG_KP_GPIO1);      //Write Keypad/GPIO SELECTION Registers
  write3Bytes((uint32_t)_PUR, REG_GPIO_PULL1);    //Write Pull-Up RESISTOR Registers
}

void SSKeypadClass::dumpreg(void) {
  Serial.println("DUMPING REGISTERS");
  uint8_t data;
  for (int x = 0x01; x < 0x2F; x++) {
    readByte(&data, x);
    Serial.print(x, HEX);
    Serial.print(":");
    Serial.print(data, HEX);
    Serial.print(" ");
  }
  Serial.print("\n");
}

uint8_t SSKeypadClass::getInterruptStatus(void) {
  uint8_t status;

  readByte(&status, REG_INT_STAT);
  return (status & 0x0F);
}

void SSKeypadClass::clearInterruptStatus(uint8_t flags) {
  flags &= 0x1F;
  writeByte(flags, REG_INT_STAT);
}

void SSKeypadClass::clearInterruptStatus(void) {
  clearInterruptStatus(0x1F);
}

uint8_t SSKeypadClass::getKeyEvent(uint8_t event) {
  uint8_t keycode;

  if (event > 9)
    return 0x00;

  readByte(&keycode, (REG_KEY_EVENT_A + event));
  return (keycode);
}

uint8_t SSKeypadClass::getKeyEvent(void) {
  return (getKeyEvent(0));
}

uint8_t SSKeypadClass::getKeyEventCount(void) {
  uint8_t count;

  readByte(&count, REG_KEY_LCK_EC);
  return (count & 0x0F);
}

uint32_t SSKeypadClass::getGPIOInterrupt(void) {
  uint32_t Ints;

  union {
    uint32_t val;
    uint8_t arr[4];
  } IntU;

  readByte(&IntU.arr[2], REG_GPIO_INT_STAT3);
  readByte(&IntU.arr[1], REG_GPIO_INT_STAT2);
  readByte(&IntU.arr[0], REG_GPIO_INT_STAT1);

  Ints = IntU.val;
  return (Ints);
}

bool SSKeypadClass::isKeyDown(char key) {
  uint8_t keycode;
  switch (key) {
    default: break;

    //Patterns
    case 0: keycode = 0x01; break;
    case 1: keycode = 0x02; break;
    case 2: keycode = 0x03; break;
    case 3: keycode = 0x04; break;
    case 4: keycode = 0x05; break;
    case 5: keycode = 0x0B; break;
    case 6: keycode = 0x0C; break;
    case 7: keycode = 0x0D; break;
    case 8: keycode = 0x0E; break;
    case 9: keycode = 0x0F; break;
    case 10: keycode = 0x15; break;
    case 11: keycode = 0x16; break;
    case 12: keycode = 0x17; break;
    case 13: keycode = 0x18; break;
    case 14: keycode = 0x19; break;
    case 15: keycode = 0x1F; break;

    //Tracks
    case 16: keycode = 0x20; break;
    case 17: keycode = 0x21; break;
    case 18: keycode = 0x22; break;
    case 19: keycode = 0x23; break;

    //Fn
    case 20: keycode = 0x29; break;

    //Play
    case 21: keycode = 0x2A; break;

    //Encoder
    case 22: keycode = 0x2B; break;

    //Save
    case 23: keycode = 0x2D; break;

    //Load
    case 24: keycode = 0x2C; break;
  }

  if (keycode & 0x80)
    return true;
  else
    return false;
}

bool SSKeypadClass::getKey(uint8_t *key) {
  uint8_t tmpkey;

  tmpkey = readKeypad();
  *key = *key & 0x7F;

  return (isKeyDown(tmpkey));
}

uint8_t SSKeypadClass::getKey(void) {
  return (readKeypad() & 0x7F);
}

SSKeypadClass SSKeypad;
